@@ -15,21 +15,21 @@
 #
 
 from __future__ import absolute_import
-
 import collections
-from itertools import chain
-from itertools import product
 import re
-import types
 import unicodedata
 
+from itertools import chain
+from itertools import product
+from builtins import object, range
 from richenum import OrderedRichEnum
 from richenum import OrderedRichEnumValue
 
+import normalize.exc as exc
 from normalize.property import SafeProperty
 from normalize.coll import Collection
+from normalize.coll import DictCollection
 from normalize.coll import ListCollection
-import normalize.exc as exc
 from normalize.record import Record
 from normalize.record import record_id
 from normalize.selector import FieldSelector
@@ -48,11 +48,12 @@ class DiffTypes(OrderedRichEnum):
     ADDED = EnumValue(2, "added", "ADDED")
     REMOVED = EnumValue(3, "removed", "REMOVED")
     MODIFIED = EnumValue(4, "modified", "MODIFIED")
+    MOVED = EnumValue(5, "moved", "MOVED")
 
 
 def _coerce_diff(dt):
     if not isinstance(dt, OrderedRichEnumValue):
-        if isinstance(dt, (int, long)):
+        if isinstance(dt, int):
             dt = DiffTypes.from_index(dt)
         else:
             dt = DiffTypes.from_canonical(dt)
@@ -67,6 +68,7 @@ class DiffInfo(Record):
     diff_type = SafeProperty(
         coerce=_coerce_diff,
         isa=DiffTypes.EnumValue,
+        empty_attr=None,
         required=True,
         doc="Enumeration describing the type of difference; a "
             ":py:class:`DiffType` value.")
@@ -123,9 +125,10 @@ class DiffOptions(object):
 
     def __init__(self, ignore_ws=True, ignore_case=False,
                  unicode_normal=True, unchanged=False,
-                 ignore_empty_slots=False,
+                 ignore_empty_slots=False, ignore_empty_items=False,
                  duck_type=False, extraneous=False,
-                 compare_filter=None, fuzzy_match=True):
+                 compare_filter=None, fuzzy_match=True, moved=False,
+                 recurse=False):
         """Create a new ``DiffOptions`` instance.
 
         args:
@@ -146,11 +149,22 @@ class DiffOptions(object):
                 those which found a difference.  Defaults to False.  Useful for
                 testing.
 
+            ``moved=``\ *BOOL*
+                Yields ``DiffInfo`` objects for comparisons where the values
+                matched, but keys or indexes were different.  Defaults to
+                False.
+
             ``ignore_empty_slots=``\ *BOOL*
                 If true, slots containing typical 'empty' values (by default,
                 just ``''`` and ``None``) are treated as if they were not set.
                 False by default.
 
+            ``ignore_empty_items=``\ *BOOL*
+                If true, items are considered to be absent from collections if
+                they have all ``None``, not set, or ``''`` in their primary key
+                fields (all compared fields in the absence of a primary key
+                definition).  False by default.
+
             ``duck_type=``\ *BOOL*
                 Normally, types must match or the result will always be
                 :py:attr:`normalize.diff.DiffTypes.MODIFIED` and the comparison
@@ -169,16 +183,24 @@ class DiffOptions(object):
                 Restrict comparison to the fields described by the passed
                 :py:class:`MultiFieldSelector` (or list of FieldSelector
                 lists/objects)
+
+            ``recurse=``\ *BOOL* During diff operations, do a deeper
+                comparison via recursion. This may be potentially very
+                expensive computationally if your records are large or
+                very nested.
         """
         self.ignore_ws = ignore_ws
         self.ignore_case = ignore_case
         self.ignore_empty_slots = ignore_empty_slots
+        self.ignore_empty_items = ignore_empty_items
         self.unicode_normal = unicode_normal
         self.fuzzy_match = fuzzy_match
         self.unchanged = unchanged
+        self.moved = moved
         self.duck_type = duck_type
         self.extraneous = extraneous
-        if isinstance(compare_filter, (MultiFieldSelector, types.NoneType)):
+        self.recurse = recurse
+        if isinstance(compare_filter, (MultiFieldSelector, type(None))):
             self.compare_filter = compare_filter
         else:
             self.compare_filter = MultiFieldSelector(*compare_filter)
@@ -190,7 +212,7 @@ class DiffOptions(object):
 
     def normalize_whitespace(self, value):
         """Normalizes whitespace; called if ``ignore_ws`` is true."""
-        if isinstance(value, unicode):
+        if isinstance(value, str):
             return u" ".join(
                 x for x in re.split(r'\s+', value, flags=re.UNICODE) if
                 len(x)
@@ -201,7 +223,7 @@ class DiffOptions(object):
     def normalize_unf(self, value):
         """Normalizes Unicode Normal Form (to NFC); called if
         ``unicode_normal`` is true."""
-        if isinstance(value, unicode):
+        if isinstance(value, str):
             return unicodedata.normalize('NFC', value)
         else:
             return value
@@ -219,7 +241,7 @@ class DiffOptions(object):
         as not specified.  Called if ``ignore_empty_slots`` is true.  Checking
         the value for emptiness happens *after* all other normalization.
         """
-        return (not value and isinstance(value, (basestring, types.NoneType)))
+        return (not value and isinstance(value, (str, type(None))))
 
     def normalize_text(self, value):
         """This hook is called by :py:meth:`DiffOptions.normalize_val` if the
@@ -239,7 +261,7 @@ class DiffOptions(object):
         return the scrubbed value or ``self._nothing`` to indicate that the
         value is not set.
         """
-        if isinstance(value, basestring):
+        if isinstance(value, str):
             value = self.normalize_text(value)
         if self.ignore_empty_slots and self.value_is_empty(value):
             value = _nothing
@@ -308,21 +330,29 @@ class DiffOptions(object):
         pk = record_id(record, type_, selector, self.normalize_object_slot)
         return pk
 
-    def id_args(self, type_, fs):
+    def id_args(self, type_, fs):  # XXX deprecated
         options = dict()
         if self.duck_type:
             options['type_'] = type_
         if self.compare_filter:
             if len(fs):
-                options['selector'] = self.compare_filter[fs][any]
+                coll_filter = self.compare_filter[fs]
             else:
-                options['selector'] = self.compare_filter[any]
+                coll_filter = self.compare_filter
+            if not coll_filter.has_none:
+                def _options(k):
+                    options['selector'] = coll_filter[k]
+                    return options
+
+                return _options
+            else:
+                options['selector'] = coll_filter[any]
         return options
 
     def is_filtered(self, prop, fs):
         if not self.extraneous and prop.extraneous:
             return True
-        return self.compare_filter and fs not in self.compare_filter
+        return self.compare_filter and not self.compare_filter[fs]
 
 
 def compare_record_iter(a, b, fs_a=None, fs_b=None, options=None):
@@ -404,19 +434,15 @@ def compare_record_iter(a, b, fs_a=None, fs_b=None, options=None):
             if one_side_nothing:
                 diff_types_found = set()
 
-            for type_union, func in COMPARE_FUNCTIONS.iteritems():
-                if isinstance(propval_a, type_union) or one_side_nothing and (
-                    isinstance(propval_b, type_union)
-                ):
-                    for diff in func(
-                        propval_a, propval_b, prop_fs_a,
-                        prop_fs_b, options,
-                    ):
-                        if one_side_nothing:
-                            if diff.diff_type != DiffTypes.NO_CHANGE:
-                                diff_types_found.add(diff.diff_type)
-                        else:
-                            yield diff
+            for diff in _diff_iter(
+                propval_a, propval_b, prop_fs_a,
+                prop_fs_b, options,
+            ):
+                if one_side_nothing:
+                    if diff.diff_type != DiffTypes.NO_CHANGE:
+                        diff_types_found.add(diff.diff_type)
+                else:
+                    yield diff
 
             if one_side_nothing:
                 net_diff = None
@@ -476,8 +502,16 @@ def collection_generator(collection):
 
     elif hasattr(collection, "itertuples"):
         return collection.itertuples()
+    elif hasattr(collection, "keys"):
+
+        def generator():
+            for key in list(collection.keys()):
+                yield (key, collection[key])
+
+    elif hasattr(collection, "items"):
+        return list(collection.items())
     elif hasattr(collection, "iteritems"):
-        return collection.iteritems()
+        return iter(collection.items())
     elif hasattr(collection, "__getitem__"):
 
         def generator():
@@ -502,6 +536,13 @@ def _nested_falsy(x):
         return x is _nothing or not x
 
 
+def _nested_empty(x):
+    if isinstance(x, tuple):
+        return all(_nested_empty(y) for y in x)
+    else:
+        return x is _nothing or x is None or x == ''
+
+
 def _fuzzy_match(set_a, set_b):
     seen = dict()
     scores = list()
@@ -577,8 +618,27 @@ def compare_collection_iter(propval_a, propval_b, fs_a=None, fs_b=None,
         type(propval_a) if propval_a is not _nothing else type(propval_b)
     )
     force_descent = (propval_a is _nothing) or (propval_b is _nothing)
+    if isinstance(coll_type.itemtype, tuple):
+        raise exc.CollectionItemTypeUnsupported()
+    elif not issubclass(coll_type.itemtype, Record):
+        iter_func = (
+            compare_dict_iter if issubclass(coll_type, DictCollection) else
+            compare_list_iter if issubclass(coll_type, ListCollection) else
+            None
+        )
+        if not iter_func:
+            raise exc.CollectionDiffUnsupported(
+                coll_type=coll_type,
+                item_type=coll_type.itemtype,
+                item_type_name=coll_type.itemtype.__name__,
+            )
+        for diff in iter_func(propval_a, propval_b, fs_a, fs_b, options):
+            yield diff
+        return
+
     id_args = options.id_args(coll_type.itemtype, fs_a)
-    if 'selector' in id_args and not id_args['selector']:
+    if not callable(id_args) and 'selector' in id_args and \
+            not id_args['selector']:
         # early exit shortcut
         return
 
@@ -590,7 +650,13 @@ def compare_collection_iter(propval_a, propval_b, fs_a=None, fs_b=None,
         seen = collections.Counter()
 
         for k, v in collection_generator(propval_x):
-            pk = options.record_id(v, **id_args)
+            if callable(id_args):
+                if fs_a + [k] not in options.compare_filter:
+                    continue
+            pk = options.record_id(
+                v, **(id_args(k) if callable(id_args) else id_args))
+            if options.ignore_empty_items and _nested_empty(pk):
+                continue
             if compare_values is None:
                 # the primary key being a tuple is taken to imply that
                 # the value type is a Record, and hence descent is
@@ -600,89 +666,127 @@ def compare_collection_iter(propval_a, propval_b, fs_a=None, fs_b=None,
             rev_key[(pk, seen[pk])] = k
             seen[pk] += 1
 
-    removed = values['a'] - values['b']
-    added = values['b'] - values['a']
-    common = values['a'].intersection(values['b'])
+    if options.recurse:
+        # we can be sure that both records have these keys
+        set_a = set(rev_keys["a"].values())
+        set_b = set(rev_keys["b"].values())
+        shared_keys = set_a.intersection(set_b)
+        removed = set_a - set_b
+        added = set_b - set_a
+        for key in shared_keys:
+            if (isinstance(propval_a, collections.abc.Iterable) and
+               isinstance(propval_b, collections.abc.Iterable)):
+                diffs = _diff_iter(propval_a[key], propval_b[key],
+                                   fs_a + [key], fs_b + [key], options)
 
-    if compare_values or force_descent:
-        descendable = (removed | added) if force_descent else common
-
-        for pk, seq in descendable:
-            if not force_descent or propval_a is not _nothing:
-                a_key = rev_keys['a'][pk, seq]
-                a_val = propval_a[a_key]
-            if not force_descent or propval_b is not _nothing:
-                b_key = rev_keys['b'][pk, seq]
-                b_val = propval_b[b_key]
-            if force_descent:
-                if propval_a is _nothing:
-                    a_key = b_key
-                    a_val = _nothing
-                else:
-                    b_key = a_key
-                    b_val = _nothing
-            selector_a = fs_a + a_key
-            selector_b = fs_b + b_key
-            # FIXME: collections of collections!
-            for diff in compare_record_iter(
-                a_val, b_val, selector_a, selector_b, options,
-            ):
-                yield diff
-
-        if not force_descent and options.fuzzy_match:
-            for a_pk_seq, b_pk_seq in _fuzzy_match(removed, added):
-                removed.remove(a_pk_seq)
-                added.remove(b_pk_seq)
-                a_key = rev_keys['a'][a_pk_seq]
-                a_val = propval_a[a_key]
-                b_key = rev_keys['b'][b_pk_seq]
-                b_val = propval_b[b_key]
-                selector_a = fs_a + a_key
-                selector_b = fs_b + b_key
-                any_diffs = False
-                for diff in compare_record_iter(
-                    a_val, b_val, selector_a, selector_b, options,
-                ):
-                    if diff.diff_type != DiffTypes.NO_CHANGE:
-                        any_diffs = True
+                for diff in diffs:
                     yield diff
 
-                if options.unchanged and not any_diffs:
+        for key in removed:
+            yield DiffInfo(diff_type=DiffTypes.REMOVED,
+                           base=fs_a + [key],
+                           other=fs_b)
+
+        for key in added:
+            yield DiffInfo(diff_type=DiffTypes.ADDED,
+                           base=fs_a,
+                           other=fs_b + [key])
+    else:
+        removed = values['a'] - values['b']
+        added = values['b'] - values['a']
+        common = values['a'].intersection(values['b'])
+
+        if compare_values or force_descent:
+            descendable = (removed | added) if force_descent else common
+
+            for pk, seq in descendable:
+                if not force_descent or propval_a is not _nothing:
+                    a_key = rev_keys['a'][pk, seq]
+                    a_val = propval_a[a_key]
+                if not force_descent or propval_b is not _nothing:
+                    b_key = rev_keys['b'][pk, seq]
+                    b_val = propval_b[b_key]
+                if force_descent:
+                    if propval_a is _nothing:
+                        a_key = b_key
+                        a_val = _nothing
+                    else:
+                        b_key = a_key
+                        b_val = _nothing
+                selector_a = fs_a + a_key
+                selector_b = fs_b + b_key
+                for diff in _diff_iter(
+                    a_val, b_val, selector_a, selector_b, options,
+                ):
+                    yield diff
+
+            if not force_descent and options.fuzzy_match:
+                for a_pk_seq, b_pk_seq in _fuzzy_match(removed, added):
+                    removed.remove(a_pk_seq)
+                    added.remove(b_pk_seq)
+                    a_key = rev_keys['a'][a_pk_seq]
+                    a_val = propval_a[a_key]
+                    b_key = rev_keys['b'][b_pk_seq]
+                    b_val = propval_b[b_key]
+                    selector_a = fs_a + a_key
+                    selector_b = fs_b + b_key
+                    any_diffs = False
+                    for diff in _diff_iter(
+                        a_val, b_val, selector_a, selector_b, options,
+                    ):
+                        if diff.diff_type != DiffTypes.NO_CHANGE:
+                            any_diffs = True
+                        yield diff
+
+                    if options.moved and a_key != b_key:
+                        yield DiffInfo(
+                            diff_type=DiffTypes.MOVED,
+                            base=fs_a + [a_key],
+                            other=fs_b + [b_key],
+                        )
+                    elif options.unchanged and not any_diffs:
+                        yield DiffInfo(
+                            diff_type=DiffTypes.NO_CHANGE,
+                            base=fs_a + [a_key],
+                            other=fs_b + [b_key],
+                        )
+
+        if options.unchanged or options.moved:
+            unchanged = values['a'] & values['b']
+            for pk, seq in unchanged:
+                a_key = rev_keys['a'][pk, seq]
+                b_key = rev_keys['b'][pk, seq]
+                if options.moved and a_key != b_key:
+                    yield DiffInfo(
+                        diff_type=DiffTypes.MOVED,
+                        base=fs_a + [a_key],
+                        other=fs_b + [b_key],
+                    )
+                elif options.unchanged:
                     yield DiffInfo(
                         diff_type=DiffTypes.NO_CHANGE,
                         base=fs_a + [a_key],
                         other=fs_b + [b_key],
                     )
 
-    if options.unchanged:
-        unchanged = values['a'] & values['b']
-        for pk, seq in unchanged:
-            a_key = rev_keys['a'][pk, seq]
-            b_key = rev_keys['b'][pk, seq]
-            yield DiffInfo(
-                diff_type=DiffTypes.NO_CHANGE,
-                base=fs_a + [a_key],
-                other=fs_b + [b_key],
-            )
+        if not force_descent:
+            for pk, seq in removed:
+                a_key = rev_keys['a'][pk, seq]
+                selector = fs_a + [a_key]
+                yield DiffInfo(
+                    diff_type=DiffTypes.REMOVED,
+                    base=selector,
+                    other=fs_b,
+                )
 
-    if not force_descent:
-        for pk, seq in removed:
-            a_key = rev_keys['a'][pk, seq]
-            selector = fs_a + [a_key]
-            yield DiffInfo(
-                diff_type=DiffTypes.REMOVED,
-                base=selector,
-                other=fs_b,
-            )
-
-        for pk, seq in added:
-            b_key = rev_keys['b'][pk, seq]
-            selector = fs_b + [b_key]
-            yield DiffInfo(
-                diff_type=DiffTypes.ADDED,
-                base=fs_a,
-                other=selector,
-            )
+            for pk, seq in added:
+                b_key = rev_keys['b'][pk, seq]
+                selector = fs_b + [b_key]
+                yield DiffInfo(
+                    diff_type=DiffTypes.ADDED,
+                    base=fs_a,
+                    other=selector,
+                )
 
 
 def compare_list_iter(propval_a, propval_b, fs_a=None, fs_b=None,
@@ -718,19 +822,32 @@ def compare_list_iter(propval_a, propval_b, fs_a=None, fs_b=None,
     removed = values['a'] - values['b']
     added = values['b'] - values['a']
 
-    if options.unchanged:
+    if options.unchanged or options.moved:
         unchanged = values['a'] & values['b']
         for v, seq in unchanged:
             a_idx = indices['a'][v, seq]
             b_idx = indices['b'][v, seq]
-            yield DiffInfo(
-                diff_type=DiffTypes.NO_CHANGE,
-                base=fs_a + [a_idx],
-                other=fs_b + [b_idx],
-            )
+            if options.moved and a_idx != b_idx:
+                yield DiffInfo(
+                    diff_type=DiffTypes.MOVED,
+                    base=fs_a + [a_idx],
+                    other=fs_b + [b_idx],
+                )
+            elif options.unchanged:
+                yield DiffInfo(
+                    diff_type=DiffTypes.NO_CHANGE,
+                    base=fs_a + [a_idx],
+                    other=fs_b + [b_idx],
+                )
+
+    removed_idx = set(indices['a'][v, seq] for v, seq in removed)
+    added_idx = set(indices['b'][v, seq] for v, seq in added)
+    modified_idx = set(removed_idx.intersection(added_idx))
 
     for v, seq in removed:
         a_key = indices['a'][v, seq]
+        if a_key in modified_idx:
+            continue
         selector = fs_a + [a_key]
         yield DiffInfo(
             diff_type=DiffTypes.REMOVED,
@@ -740,6 +857,8 @@ def compare_list_iter(propval_a, propval_b, fs_a=None, fs_b=None,
 
     for v, seq in added:
         b_key = indices['b'][v, seq]
+        if b_key in modified_idx:
+            continue
         selector = fs_b + [b_key]
         yield DiffInfo(
             diff_type=DiffTypes.ADDED,
@@ -747,6 +866,13 @@ def compare_list_iter(propval_a, propval_b, fs_a=None, fs_b=None,
             other=selector,
         )
 
+    for idx in modified_idx:
+        yield DiffInfo(
+            diff_type=DiffTypes.MODIFIED,
+            base=fs_a + [idx],
+            other=fs_b + [idx],
+        )
+
 
 def compare_dict_iter(propval_a, propval_b, fs_a=None, fs_b=None,
                       options=None):
@@ -781,19 +907,32 @@ def compare_dict_iter(propval_a, propval_b, fs_a=None, fs_b=None,
     removed = values['a'] - values['b']
     added = values['b'] - values['a']
 
-    if options.unchanged:
+    if options.moved or options.unchanged:
         unchanged = values['a'] & values['b']
         for v, seq in unchanged:
             a_key = rev_keys['a'][v, seq]
             b_key = rev_keys['b'][v, seq]
-            yield DiffInfo(
-                diff_type=DiffTypes.NO_CHANGE,
-                base=fs_a + [a_key],
-                other=fs_b + [b_key],
-            )
+            if options.moved and a_key != b_key:
+                yield DiffInfo(
+                    diff_type=DiffTypes.MOVED,
+                    base=fs_a + [a_key],
+                    other=fs_b + [b_key],
+                )
+            elif options.unchanged:
+                yield DiffInfo(
+                    diff_type=DiffTypes.NO_CHANGE,
+                    base=fs_a + [a_key],
+                    other=fs_b + [b_key],
+                )
+
+    removed_keys = set(rev_keys['a'][v, seq] for v, seq in removed)
+    added_keys = set(rev_keys['b'][v, seq] for v, seq in added)
+    modified_keys = set(removed_keys.intersection(added_keys))
 
     for v, seq in removed:
         a_key = rev_keys['a'][v, seq]
+        if a_key in modified_keys:
+            continue
         selector = fs_a + [a_key]
         yield DiffInfo(
             diff_type=DiffTypes.REMOVED,
@@ -803,6 +942,8 @@ def compare_dict_iter(propval_a, propval_b, fs_a=None, fs_b=None,
 
     for v, seq in added:
         b_key = rev_keys['b'][v, seq]
+        if b_key in modified_keys:
+            continue
         selector = fs_b + [b_key]
         yield DiffInfo(
             diff_type=DiffTypes.ADDED,
@@ -810,6 +951,13 @@ def compare_dict_iter(propval_a, propval_b, fs_a=None, fs_b=None,
             other=selector,
         )
 
+    for key in modified_keys:
+        yield DiffInfo(
+            diff_type=DiffTypes.MODIFIED,
+            base=fs_a + [key],
+            other=fs_b + [key],
+        )
+
 
 COMPARE_FUNCTIONS = {
     list: compare_list_iter,
@@ -850,11 +998,18 @@ def diff_iter(base, other, options=None, **kwargs):
     elif len(kwargs):
         raise exc.DiffOptionsException()
 
+    null_fs = FieldSelector(tuple())
+    return _diff_iter(base, other, null_fs, null_fs, options)
+
+
+def _diff_iter(base, other, fs_a, fs_b, options):
     generators = []
 
-    for type_union, func in COMPARE_FUNCTIONS.iteritems():
-        if isinstance(base, type_union):
-            generators.append(func(base, other, options=options))
+    for type_union, func in COMPARE_FUNCTIONS.items():
+        matches = (isinstance(base, type_union) if base is not _nothing else
+                   isinstance(other, type_union))
+        if matches:
+            generators.append(func(base, other, fs_a, fs_b, options=options))
 
     if len(generators) == 1:
         return generators[0]
@@ -894,7 +1049,7 @@ class Diff(ListCollection):
                 diffstate['==X'].append(diff.base)
 
         prefix_paths = []
-        for k, v in diffstate.items():
+        for k, v in list(diffstate.items()):
             prefix_paths.append(
                 "{prefix}({paths})".format(
                     prefix=k,
@@ -910,7 +1065,7 @@ class Diff(ListCollection):
                     "{prefix}({paths})".format(
                         prefix=k,
                         paths=MultiFieldSelector(*v).path,
-                    ) for (k, v) in diffstate.items()
+                    ) for (k, v) in list(diffstate.items())
                 ) if diffstate else ""
             ),
         )
