@@ -14,6 +14,7 @@
 # http://github.com/hearsaycorp/normalize
 #
 
+from builtins import str
 import normalize.coll
 import normalize.exc as exc
 import normalize.record
@@ -27,12 +28,24 @@ def record_id(object_, type_=None, selector=None, normalize_object_slot=None):
     Curiously, this function resembles conversion between a "record" and a
     "tuple": stripping the logical names from the atomic values.
     """
-    if type_ is None:
+    if type_ is None or isinstance(type_, tuple):
         type_ = type(object_)
 
     key_vals = list()
-    pk_cols = type_.primary_key
-    if selector and pk_cols and not all(x.name in selector for x in pk_cols):
+    if hasattr(type_, "primary_key"):
+        pk_cols = type_.primary_key
+    elif object_.__hash__:
+        return object_
+    else:
+        raise exc.IdentityCrisis(
+            val=object_,
+            val_repr=repr(object_),
+            val_type=type_,
+            val_type_name=type_.__name__,
+        )
+    if selector and pk_cols and not all(
+        selector[(x.name,)] for x in pk_cols
+    ):
         pk_cols = None
 
     if not pk_cols and issubclass(
@@ -46,7 +59,7 @@ def record_id(object_, type_=None, selector=None, normalize_object_slot=None):
         return tuple(
             record_id(
                 v, type_.itemtype, selector[k], normalize_object_slot,
-            ) for k, v in gen if k in selector
+            ) for k, v in gen if selector[(k,)]
         ) if selector else tuple(
             record_id(v, type_.itemtype, None, normalize_object_slot) for
             k, v in gen
@@ -56,7 +69,7 @@ def record_id(object_, type_=None, selector=None, normalize_object_slot=None):
         all_properties = type_._sorted_properties
         if selector:
             all_properties = tuple(
-                x for x in all_properties if x.name in selector
+                x for x in all_properties if selector[(x.name,)]
             )
 
     for prop in pk_cols or all_properties:
