@@ -30,7 +30,9 @@ import inspect
 import warnings
 import weakref
 
+import normalize.empty as empty
 import normalize.exc as exc
+from normalize.property.meta import looks_like_v1_none
 from normalize.property.meta import MetaProperty
 
 
@@ -42,17 +44,17 @@ class _Default(object):
 _none = _Default()
 
 
-class Property(object):
+class Property(metaclass=MetaProperty):
     """This is the base class for all property types.  It is a data descriptor,
     so care should be taken before adding any ``SPECIALMETHODS`` which might
     change the way it behaves.
     """
-    __metaclass__ = MetaProperty
     __safe_unless_ro__ = False
 
-    def __init__(self, isa=None,  coerce=None, check=None,
+    def __init__(self, isa=None, coerce=None, check=None,
                  required=False, default=_none, traits=None,
-                 extraneous=False, doc=None):
+                 extraneous=False, empty_attr=_none,
+                 doc=None):
         """Declares a new standard Property.  Note: if you pass arguments which
         are not understood by this constructor, or pass extra property traits
         to ``traits``, then the call will be redirected to a sub-class; see
@@ -94,6 +96,14 @@ class Property(object):
                 is ``["safe"]``, and any unknown keyword arguments will add
                 extra traits on.
 
+            ``empty_attr=``\ *METHODNAME*
+                (deprecated) Specify an auxiliary method name which
+                returns the value if the attribute is set, otherwise
+                an ``empty`` proxy value.  Defaults to the name of the
+                attribute with a ``0`` appended, or ``None`` if the
+                attribute already ends with a number (disabling the
+                accessor)
+
             ``extraneous=``\ *BOOL*
                 This Property is considered *denormalized* and does not affect
                 the ``Record`` equality operator.  Visitor functions typically
@@ -102,6 +112,7 @@ class Property(object):
 
             ``doc=``\ *STR*
                 Specify a docstring for the property.
+
         """
         self.name = None
         self.class_ = None
@@ -142,10 +153,11 @@ class Property(object):
         self.coerce = coerce or isa
         if self.coerce and not self.valuetype:
             raise exc.CoerceWithoutType()
+        self.empty_attr = empty_attr
         self.extraneous = extraneous
 
     def func_info(self, func):
-        args = inspect.getargspec(func)
+        args = inspect.getfullargspec(func)
         is_method = False
         if not args.args:
             required_args = 0
@@ -162,8 +174,14 @@ class Property(object):
     def bound(self):
         return bool(self.class_)
 
-    def bind(self, class_, name):
+    def set_name(self, name):
         self.name = name
+        if self.empty_attr is _none:
+            self.empty_attr = (
+                (name + "0") if name[-1] not in "0123456789" else None
+            )
+
+    def bind(self, class_):
         self.class_ = weakref.ref(class_)
 
     @property
@@ -171,7 +189,10 @@ class Property(object):
         """Returns the name of the ``Record`` class this ``Property`` is
         attached to, and attribute name it is attached as."""
         if not self.bound:
-            return "(unbound)"
+            if self.name is not None:
+                return "(unbound).%s" % self.name
+            else:
+                return "(unbound)"
         elif not self.class_():
             classname = "(GC'd class)"
         else:
@@ -180,14 +201,18 @@ class Property(object):
 
     def type_safe_value(self, value, _none_ok=False):
         if value is None and self.required and not self.valuetype:
-            raise ValueError("%s is required" % self.fullname)
+            raise exc.PropertyRequired(prop=self)
         if self.valuetype and not isinstance(value, self.valuetype):
             try:
                 new_value = self.coerce(value)
+            except exc.SubtypeCoerceError as e:
+                # this particular coerce error will be re-caught below,
+                # unless the coerce method returned None
+                new_value = e.coerced
             except Exception as e:
                 raise exc.CoerceError(
-                    prop=self.fullname,
-                    value=repr(value),
+                    prop=self,
+                    passed=value,
                     exc=e,
                     func=(
                         "%s constructor" % self.coerce.__name__ if
@@ -207,15 +232,16 @@ class Property(object):
                     return _none
                 else:
                     raise exc.ValueCoercionError(
-                        prop=self.fullname,
-                        value=repr(value),
-                        coerced=repr(new_value),
+                        prop=self,
+                        passed=value,
+                        coerced=new_value,
                     )
             else:
                 value = new_value
         if self.check and not self.check(value):
-            raise ValueError(
-                "%s value '%r' failed type check" % (self.fullname, value)
+            raise exc.ValueCheckError(
+                prop=self,
+                passed=value,
             )
         return value
 
@@ -241,26 +267,59 @@ class Property(object):
 
         if new_value is _none:
             if self.required:
-                raise ValueError("%s is required" % self.fullname)
+                raise exc.PropertyRequired(prop=self)
         else:
             obj.__dict__[self.name] = new_value
 
     def eager_init(self):
         return self.required or self.default is not _none
 
+    def attribute_error_hook(self):
+        raise AttributeError(self.fullname)
+
     def __get__(self, obj, type_=None):
         """Default getter; does NOT fall back to regular descriptor behavior
         """
         if obj is None:
             return self
         if self.name not in obj.__dict__:
-            raise AttributeError(self.fullname)
+            return self.attribute_error_hook()
         return obj.__dict__[self.name]
 
+    def slot_is_empty(self, obj):
+        return self.name not in obj.__dict__
+
     def __str__(self):
         metaclass = str(type(self).__name__)
         return "<%s %s>" % (metaclass, self.fullname)
 
+    def aux_props(self):
+        """This method is available for property traits to provide extra class
+        attributes which are added to the class they are defined in during
+        class creation.  The default implementation is responsible for defining
+        ``empty_attr`` attributes.
+
+        The return value should be an iterable list of 2-tuples, with the first
+        member of each 2-tuple being the attribute name and the second being
+        the value to insert.
+        """
+        if self.empty_attr is not None:
+            return ((self.empty_attr, EmptyAuxProp(self)), )
+        else:
+            return ()
+
+
+class EmptyAuxProp(object):
+    def __init__(self, prop):
+        self.prop = prop
+        self.valuetype = prop.valuetype or any
+
+    def __get__(self, obj, type_=None):
+        try:
+            return self.prop.__get__(obj)
+        except AttributeError:
+            return empty.placeholder(self.valuetype)
+
 
 class LazyProperty(Property):
     """This declares a property which has late evaluation using its 'default'
@@ -309,6 +368,9 @@ class LazyProperty(Property):
 
         return super(LazyProperty, self).__get__(obj, type_)
 
+    def slot_is_empty(self, obj):
+        return False
+
 
 class ROProperty(Property):
     """A read-only property throws an exception when the attribute slot is
@@ -354,7 +416,7 @@ class SafeProperty(Property):
         """Checks the property's ``required`` setting, and allows the delete if
         it is false"""
         if self.required:
-            raise ValueError("%s is required" % self.fullname)
+            raise exc.PropertyRequired(prop=self)
         del obj.__dict__[self.name]
 
 
@@ -369,6 +431,46 @@ class LazySafeProperty(SafeProperty, LazyProperty):
         return super(LazySafeProperty, self).__get__(obj, type_)
 
 
+class V1Property(SafeProperty):
+    __trait__ = "v1"
+
+    def __init__(self, v1_none=_none, **kwargs):
+        """Passing 'v1_none=None' or declaring with 'V1Property' will
+        suppress AttributeError when the attribute is not set.
+        """
+        if not v1_none.__hash__:
+            raise exc.NoneMutable(passed=v1_none)
+        self.v1_upgraded = False
+        if v1_none is _none and 'default' in kwargs and \
+           looks_like_v1_none(kwargs['default']):
+            v1_none = kwargs['default']
+            del kwargs['default']
+            self.v1_upgraded = True
+        super(V1Property, self).__init__(**kwargs)
+        self.v1_none = None if v1_none is _none else v1_none
+        if self.empty_attr is _none and not self.v1_upgraded:
+            self.empty_attr = None
+
+    def slot_is_empty(self, obj):
+        if self.v1_upgraded:
+            return False
+        return super(V1Property, obj).slot_is_empty(obj)
+
+    def attribute_error_hook(self):
+        return self.v1_none
+
+    def __set__(self, obj, value):
+        """This setter checks the type of the value before allowing it to be
+        set."""
+        try:
+            super().__set__(obj, value)
+        except exc.CoercionError:
+            if value is None or value == self.v1_none:
+                self.__delete__(obj)
+            else:
+                raise
+
+
 class DiffasProperty(Property):
     __trait__ = "diffas"
 
@@ -401,9 +503,7 @@ def make_property_type(name, base_type=Property,
 
     The typical use of this function is to make types for the API you are
     mapping so that, for instance, any time they use a date you can convert
-    it in a consistent way to a ``datetime.date``, or to supply
-    ``default=None`` because you and prefer subtle bugs caused by stray
-    ``None`` values to ``AttributeError`` exceptions.
+    it in a consistent way to a ``datetime.date``.
 
     It's also used by :py:mod:`normalize.property.types` to create all
     of its Property subclasses.
@@ -450,7 +550,7 @@ def make_property_type(name, base_type=Property,
         trait_name = "trait%d" % trait_num
 
     def __init__(self, **kwargs):
-        for arg, val in default_kwargs.iteritems():
+        for arg, val in default_kwargs.items():
             if arg not in kwargs:
                 kwargs[arg] = val
         return super(self_type[0], self).__init__(**kwargs)
