@@ -15,6 +15,7 @@
 #
 
 
+from normalize.coll import DictCollection
 from normalize.coll import ListCollection
 from normalize.coll import _make_generic
 import normalize.exc as exc
@@ -68,13 +69,14 @@ class SafeCollectionProperty(CollectionProperty, SafeProperty):
 
 class ListProperty(CollectionProperty):
     __trait__ = "list"
+    coll_type = ListCollection
 
     def __init__(self, list_of=None, **kwargs):
         if list_of is None:
             list_of = kwargs.pop("of", None)
         if not list_of:
             raise exc.ListOfWhat()
-        colltype = kwargs.pop('coll', ListCollection)
+        colltype = kwargs.pop('coll', self.coll_type)
         if not issubclass(colltype, ListCollection):
             raise exc.ListPropertyMustDeriveListCollection(
                 got=colltype.__name__,
@@ -83,3 +85,23 @@ class ListProperty(CollectionProperty):
         super(ListProperty, self).__init__(
             of=list_of, coll=colltype, **kwargs
         )
+
+
+class DictProperty(CollectionProperty):
+    __trait__ = "dict"
+    coll_type = DictCollection
+
+    def __init__(self, dict_of=None, **kwargs):
+        if dict_of is None:
+            dict_of = kwargs.pop("of", None)
+        if not dict_of:
+            raise exc.DictOfWhat()
+        colltype = kwargs.pop('coll', self.coll_type)
+        if not issubclass(colltype, DictCollection):
+            raise exc.DictPropertyMustDeriveDictCollection(
+                got=colltype.__name__,
+            )
+
+        super(DictProperty, self).__init__(
+            of=dict_of, coll=colltype, **kwargs
+        )
