@@ -27,6 +27,12 @@ PROPERTY_TYPES = dict()
 DUCKWARGS = defaultdict(set)
 
 
+# a test for whether a value passed to 'default' is sufficient to qualify the
+# attribute for v1 upgrade
+def looks_like_v1_none(value):
+    return not value and value.__hash__ and not callable(value)
+
+
 def has(selfie, self, args, kwargs):
     """This is called 'has' but is called indirectly.  Each Property sub-class
     is installed with this function which replaces their __new__.
@@ -87,6 +93,12 @@ def has(selfie, self, args, kwargs):
     elif "ro" not in all_traits and safe_unless_ro:
         all_traits.add("safe")
 
+    if "v1" not in all_traits:
+        if 'default' in kwargs and looks_like_v1_none(kwargs['default']):
+            all_traits.add("v1")
+            if 'safe' not in all_traits:
+                all_traits.add("safe")
+
     trait_set_key = tuple(sorted(all_traits))
 
     if trait_set_key not in PROPERTY_TYPES:
@@ -104,12 +116,12 @@ def has(selfie, self, args, kwargs):
 def _merge_camel_case_names(base_name, new_name):
     import re
     name_parts = re.sub(
-        r'([a-z])([A-Z])', lambda m: "%s,%s" % m.groups(), base_name,
+        r'([a-z\d])([A-Z])', lambda m: "%s,%s" % m.groups(), base_name,
     ).split(",")
 
     other_parts = list(
         x for x in re.sub(
-            r'([a-z])([A-Z])', lambda m: "%s,%s" % m.groups(), new_name,
+            r'([a-z\d])([A-Z])', lambda m: "%s,%s" % m.groups(), new_name,
         ).split(",") if x not in name_parts
     )
 
@@ -131,12 +143,12 @@ def create_property_type_from_traits(trait_set):
     """
     wanted_traits = set(trait_set)
     stock_types = dict(
-        (k, v) for k, v in PROPERTY_TYPES.items() if
+        (k, v) for k, v in list(PROPERTY_TYPES.items()) if
         set(k).issubset(wanted_traits)
     )
 
     traits_available = set()
-    for key in stock_types.keys():
+    for key in list(stock_types.keys()):
         traits_available.update(key)
 
     missing_traits = wanted_traits - traits_available
@@ -153,7 +165,7 @@ def create_property_type_from_traits(trait_set):
         # be somewhat deterministic: always start with types which provide the
         # 'first' trait on the list
         start_with = set(
-            k for k in stock_types.keys() if k and k[0] == trait_set[0]
+            k for k in list(stock_types.keys()) if k and k[0] == trait_set[0]
         )
 
         # prefer extending already composed trait sets, by only adding to the
@@ -222,7 +234,7 @@ class MetaProperty(type):
         attrs['__new__'] = _has
         duckwargs = set()
         if '__init__' in attrs:
-            new_kwargs = inspect.getargspec(attrs['__init__']).args
+            new_kwargs = inspect.getfullargspec(attrs['__init__']).args
             if new_kwargs:
                 duckwargs.update(new_kwargs)
         traits = set()
