@@ -16,12 +16,12 @@
 
 """``normalize.property.types`` provides an assortment of pre-generated
 types"""
-
-from datetime import date
-from datetime import datetime
-from sys import maxint
+import datetime
+import numbers
+from sys import maxsize
 
 from . import make_property_type
+from ..subtype import subtype
 
 try:
     from dateutil.parser import parse as parse_datetime
@@ -39,7 +39,7 @@ except ImportError:
         datetime_stripped = not_a_datetime.replace(
             "-", "").replace("T", "").replace(" ", "")
         if len(datetime_stripped) in formats:
-            return datetime.strptime(
+            return datetime.datetime.strptime(
                 datetime_stripped, formats[len(datetime_stripped)],
             )
         else:
@@ -56,25 +56,31 @@ IntProperty = make_property_type(
     },
 )
 LongProperty = make_property_type(
-    "LongProperty", isa=long, trait_name="long",
+    "LongProperty", isa=int, trait_name="long",
     attrs={
         "__doc__": "A property which must be a ``long``",
     },
 )
 IntegerProperty = make_property_type(
-    "IntegerProperty", isa=(int, long), trait_name="integer",
-    coerce=lambda x: (
-        int(x) if abs(float(x)) < maxint else long(x)
-    ),
+    "IntegerProperty", isa=numbers.Integral, trait_name="integer",
+    coerce=lambda x: int(x),
     attrs={
-        "__doc__": "A property which may be either an int or a long",
+        "__doc__": "A property which holds an integer, int or long",
+    },
+)
+NumberProperty = make_property_type(
+    "NumberProperty", isa=numbers.Number, trait_name="number",
+    coerce=lambda x: coerce_number(x),
+    attrs={
+        "__doc__": "A property which holds a number type (eg float, int) "
+                   "with automatic cast from string",
     },
 )
 StringProperty = make_property_type(
-    "StringProperty", isa=basestring, coerce=str, trait_name="str",
+    "StringProperty", isa=str, trait_name="str",
     attrs={
-        "__doc__": "A property which must be a ``basestring``, and if "
-                   "not, it is coerced using ``str``",
+        "__doc__": "A property which must be a ``basestring`` or "
+                   "``unicode``, and if not, throws a coerce error",
     },
 )
 FloatProperty = make_property_type(
@@ -85,10 +91,12 @@ FloatProperty = make_property_type(
 )
 UnicodeProperty = make_property_type(
     "UnicodeProperty", base_type=StringProperty,
-    isa=unicode, coerce=unicode, trait_name="unicode",
+    isa=str, coerce=(lambda s: str(s) if isinstance(s, str) else s),
+    trait_name="unicode",
     attrs={
-        "__doc__": "A property which must be a ``unicode``, and if "
-                   "not, it is coerced using ``unicode``",
+        "__doc__": "A property which must be a ``unicode`` or ``str`` "
+                   "(it is upgraded to ``unicode`` if it is passed in as "
+                   "a ``str``)",
     },
 )
 
@@ -96,8 +104,8 @@ UnicodeProperty = make_property_type(
 def coerce_datetime(not_a_datetime):
     if isinstance(not_a_datetime, date):
         tt = not_a_datetime.timetuple()
-        return datetime(*(tt[0:6]))
-    elif isinstance(not_a_datetime, basestring):
+        return datetime.datetime(*(tt[0:6]))
+    elif isinstance(not_a_datetime, str):
         return parse_datetime(not_a_datetime)
     else:
         raise ValueError(
@@ -106,7 +114,7 @@ def coerce_datetime(not_a_datetime):
 
 
 def coerce_date(not_a_date):
-    if isinstance(not_a_date, datetime) or (
+    if isinstance(not_a_date, datetime.datetime) or (
         hasattr(not_a_date, "date") and callable(not_a_date.date)
     ):
         return not_a_date.date()
@@ -114,21 +122,40 @@ def coerce_date(not_a_date):
         return coerce_datetime(not_a_date).date()
 
 
+def coerce_number(not_a_number):
+    if isinstance(not_a_number, str):
+        try:
+            return int(not_a_number)
+        except ValueError:
+            return float(not_a_number)
+    else:
+        return float(not_a_number)
+
+
+date = subtype(
+    "date",
+    of=datetime.date,
+    where=lambda x: not isinstance(x, datetime.datetime),
+)
+
+
 DateProperty = make_property_type(
     "DateProperty",
     trait_name="date", isa=date, coerce=coerce_date,
+    json_out=lambda dt: dt.isoformat(),
     attrs={
-        "__doc__": "A property which must hold a python date (or "
-                   "datetime); coercion from string is provided via "
-                   "``dateutil.parse``.  Note that datetime is a date "
-                   "subclass, and supports all its methods, so a "
-                   "DateProperty might contain a datetime instance.",
+        "__doc__": "A property which must hold a python date; coercion "
+                   "from string is provided via ``dateutil.parse``.  "
+                   "As of normalize v1, if a ``datetime.datetime`` "
+                   "instance is assigned to a ``DateProperty``, it will "
+                   "be truncated to a ``datetime.date``.",
     },
 )
 DatetimeProperty = make_property_type(
     "DatetimeProperty",
-    trait_name="datetime", isa=datetime,
+    trait_name="datetime", isa=datetime.datetime,
     coerce=coerce_datetime,
+    json_out=lambda dt: dt.isoformat(),
     attrs={
         "__doc__": "A property which must holds a python datetime.  "
                    "Correct timezone handling is currently TODO and "
@@ -138,4 +165,4 @@ DatetimeProperty = make_property_type(
 )
 
 
-__all__ = tuple(k for k in globals().keys() if k.endswith("Property"))
+__all__ = tuple(k for k in list(globals().keys()) if k.endswith("Property"))
