@@ -21,14 +21,13 @@ from normalize.identity import record_id
 from normalize.record.meta import RecordMeta
 
 
-class _Unset(object):
+class _Unset:
     pass
 
 
-class Record(object):
+class Record(metaclass=RecordMeta):
     """Base class for normalize instances and collections.
     """
-    __metaclass__ = RecordMeta
 
     def __init__(self, init_dict=None, **kwargs):
         """Instantiates a new ``Record`` type.
@@ -52,11 +51,12 @@ class Record(object):
             raise exc.AmbiguousConstruction()
         if not init_dict:
             init_dict = kwargs
-        for prop, val in init_dict.iteritems():
+        for prop, val in init_dict.items():
             meta_prop = type(self).properties.get(prop, None)
             if meta_prop is None:
                 raise exc.PropertyNotKnown(
                     propname=prop,
+                    recordtype=type(self),
                     typename=type(self).__name__,
                 )
             meta_prop.init_prop(self, val)
@@ -123,7 +123,7 @@ class Record(object):
         fine-tuned."""
         if type(self) != type(other):
             return False
-        for propname, prop in type(self).properties.iteritems():
+        for propname, prop in type(self).properties.items():
             if not prop.extraneous:
                 if getattr(self, propname, _Unset) != getattr(
                     other, propname, _Unset
