@@ -16,17 +16,27 @@
 
 from __future__ import absolute_import
 
+from builtins import range, object
 import collections
 from copy import deepcopy
 import functools
 import re
-import types
 
+from normalize.coll import DictCollection
 from normalize.coll import ListCollection
 from normalize.exc import FieldSelectorAttributeError
 from normalize.exc import FieldSelectorException
 from normalize.exc import FieldSelectorKeyError
-from normalize.exc import FieldSelectorOperationUnsupported
+
+
+def _try_index(instance, selector):
+    if isinstance(instance, str):
+        return False
+    if isinstance(selector, int):
+        return True
+    if getattr(instance, "__getitem__", False):
+        return True
+    return False
 
 
 @functools.total_ordering
@@ -62,8 +72,7 @@ class FieldSelector(object):
             # Validate the selector
             if any(
                 e for e in expr_selectors if not (
-                    isinstance(e, basestring) or
-                    isinstance(e, (int, long)) or e is None
+                    isinstance(e, str) or isinstance(e, int) or e is None
                 )
             ):
                 raise ValueError(
@@ -76,7 +85,7 @@ class FieldSelector(object):
     def add_property(self, prop):
         """Extends the selector, adding a new attribute property lookup at the
         end, specified by name."""
-        if not isinstance(prop, basestring):
+        if not isinstance(prop, str):
             raise ValueError(
                 "properties must be specified by their string name"
             )
@@ -85,7 +94,7 @@ class FieldSelector(object):
     def add_index(self, index):
         """Extends the selector, adding a new indexed collection lookup at the
         end."""
-        if not isinstance(index, (int, long)):
+        if not isinstance(index, int):
             raise ValueError("index must be an int or a long")
         self.selectors.append(index)
 
@@ -117,6 +126,12 @@ class FieldSelector(object):
         If there is a problem, this method will typically raise
         :py:class:`FieldSelectorException`.
 
+        If the FieldSelector contains ``None``, then exceptions from iterating
+        over all of the sub-fields are suppressed, so long as one of the items
+        succeed.  The return value will have ``None`` for the values which did
+        not have the attribute.  If all the items fail, then one of the
+        exceptions will be bubbled up to the caller.
+
         ::
 
            record.foo = "bar"
@@ -127,8 +142,22 @@ class FieldSelector(object):
         for selector in self.selectors:
             if selector is None:
                 sub_field_selector = type(self)(self.selectors[i + 1:])
-                return [sub_field_selector.get(r) for r in record]
-            elif isinstance(selector, (int, long)):
+                rv = []
+                all_exc = None
+                for r in record:
+                    val = None
+                    try:
+                        val = sub_field_selector.get(r)
+                        all_exc = False
+                    except FieldSelectorException as e:
+                        if all_exc is None:
+                            all_exc = e
+                    rv.append(val)
+                if all_exc:
+                    raise all_exc
+                else:
+                    return rv
+            elif _try_index(record, selector):
                 try:
                     record = record[selector]
                 except IndexError:
@@ -140,6 +169,18 @@ class FieldSelector(object):
             i = i + 1
         return record
 
+    def get_or_none(self, record):
+        """Evaluate the FieldSelector's path and return the value, like
+        ``.get``.
+
+        If there is a problem, such as the attribute does not exist,
+        this method will return ``None``.
+        """
+        try:
+            return self.get(record)
+        except FieldSelectorException:
+            return None
+
     def put(self, record, value):
         """
         Sets the field referenced by this field selector for the given Record
@@ -154,7 +195,7 @@ class FieldSelector(object):
             selector = self.selectors[0]
             if selector is None:
                 record[:] = value
-            elif isinstance(selector, (int, long)):
+            elif _try_index(record, selector):
                 try:
                     record[selector] = value
                 except LookupError:
@@ -177,7 +218,7 @@ class FieldSelector(object):
                 for x in record:
                     sub_selector.put(x, value)
             else:
-                if isinstance(selector, (int, long)):
+                if _try_index(record, selector):
                     try:
                         sub_record = record[selector]
                     except IndexError:
@@ -210,21 +251,38 @@ class FieldSelector(object):
            print obj.stuff[0].name  # "Bob"
         """
         i = 0
+        put_final = False
         for selector in self.selectors[:-1]:
             if selector is None:
                 sub_field_selector = type(self)(self.selectors[i + 1:])
                 return sum(sub_field_selector.post(r) for r in record)
-            elif isinstance(selector, (int, long)):
+
+            # supports the case where an item is created by setting
+            # a single required field value
+            kwarg = {}
+            if i >= len(self.selectors) - 2 and self.selectors[-1] is not None:
+                kwarg[self.selectors[-1]] = value
+
+            if _try_index(record, selector):
                 try:
                     record = record[selector]
-                except LookupError:
-                    if len(record) != selector:
-                        raise ValueError(
-                            "FieldSelector set out of order: "
-                            "[%d]" % selector
-                        )
-                    record.append(type(record).itemtype())
+                except IndexError:
+                    if isinstance(selector, int):
+                        if len(record) != selector:
+                            raise ValueError(
+                                "FieldSelector set out of order: "
+                                "[%d]" % selector
+                            )
+                        record.append(type(record).itemtype(**kwarg))
+                        if kwarg:
+                            put_final = True
+                    else:
+                        raise
                     record = record[selector]
+                except KeyError:
+                    record[selector] = type(record).itemtype(**kwarg)
+                    if kwarg:
+                        put_final = True
             else:
                 if not hasattr(record, selector):
                     prop = type(record).properties[selector]
@@ -233,10 +291,13 @@ class FieldSelector(object):
                             "Must specify default= or isa= to auto-vivify "
                             "%s" % prop
                         )
-                    setattr(record, selector, prop.valuetype())
+                    setattr(record, selector, prop.valuetype(**kwarg))
+                    if kwarg:
+                        put_final = True
                 record = getattr(record, selector)
             i = i + 1
-        type(self)([self.selectors[-1]]).put(record, value)
+        if not put_final:
+            type(self)([self.selectors[-1]]).put(record, value)
         return 1
 
     def delete(self, record):
@@ -246,6 +307,10 @@ class FieldSelector(object):
         If there is a problem, this method will typically raise
         :py:class:`FieldSelectorException`.
 
+        This function has the same behavior with respect to collections as
+        ``get()``: it will only raise an exception if all of the items in the
+        set did not have the property to delete.
+
         ::
 
            record.foo = "bar"
@@ -257,9 +322,19 @@ class FieldSelector(object):
         for selector in self.selectors[:-1]:
             if selector is None:
                 sub_field_selector = type(self)(self.selectors[i + 1:])
+                all_exc = None
                 for r in record:
-                    sub_field_selector.delete(r)
-            elif isinstance(selector, (int, long)):
+                    try:
+                        sub_field_selector.delete(r)
+                        all_exc = False
+                    except FieldSelectorException as e:
+                        if all_exc is None:
+                            all_exc = e
+                if all_exc:
+                    raise all_exc
+                else:
+                    return
+            elif _try_index(record, selector):
                 try:
                     record = record[selector]
                 except IndexError:
@@ -270,12 +345,15 @@ class FieldSelector(object):
                 record = getattr(record, selector)
             i = i + 1
         to_delete = self.selectors[-1]
-        if isinstance(to_delete, (int, long, types.NoneType)):
-            # can't do this until the coll types support mutation operations
-            raise FieldSelectorOperationUnsupported(
-                operation="delete",
-                target="collection",
-            )
+        if _try_index(record, to_delete):
+            if to_delete is None:
+                # empty out a collection
+                if hasattr(record, "clear") and callable(record.clear):
+                    record.clear()
+                else:
+                    record[:] = ()
+            else:
+                del record[to_delete]
         else:
             if hasattr(record, to_delete):
                 delattr(record, to_delete)
@@ -289,6 +367,9 @@ class FieldSelector(object):
             )
         return self.selectors == other.selectors
 
+    def __hash__(self):
+        return super(FieldSelector, self).__hash__()
+
     def __ne__(self, other):
         """
         functools.total_ordering() currently doesn't take care of __ne__().
@@ -305,7 +386,7 @@ class FieldSelector(object):
         end = len(self.selectors)
         if len(self.selectors) > len(other.selectors):
             end = len(other.selectors)
-        for i in xrange(end):
+        for i in range(end):
             self_selector = self.selectors[i]
             other_selector = other.selectors[i]
             if self_selector == other_selector:
@@ -346,9 +427,9 @@ class FieldSelector(object):
             print fs + bar  # <FieldSelector: .foo.bar>
             print fs + [0]  # <FieldSelector: .foo[0]>
         """
-        if isinstance(other, (basestring, int, long)):
+        if isinstance(other, (str, int)):
             return type(self)(self.selectors + [other])
-        elif isinstance(other, collections.Iterable):
+        elif isinstance(other, collections.abc.Iterable):
             return type(self)(self.selectors + list(other))
         elif isinstance(other, FieldSelector):
             return type(self)(self).extend(other)
@@ -419,11 +500,11 @@ class FieldSelector(object):
 
 
 def _fmt_selector_path(selector):
-    if isinstance(selector, (int, long)):
+    if isinstance(selector, int):
         return "[%d]" % selector
     elif selector is None:
         return "[*]"
-    elif re.search(r'[^a-z_]', selector):
+    elif not re.match(r'^[a-zA-Z_]\w*$', selector):
         return u"['%s']" % selector.replace("\\", "\\\\").replace("'", "\\'")
     else:
         return ".%s" % selector
@@ -479,6 +560,7 @@ class MultiFieldSelector(object):
     into a single tree structure."""
 
     FieldSelector = FieldSelector
+    _complete_mfs = None
 
     def __init__(self, *others):
         """Returns a MultiFieldSelector based on combining the passed-in
@@ -495,7 +577,7 @@ class MultiFieldSelector(object):
         heads = collections.defaultdict(set)
         for other in others:
             if isinstance(other, MultiFieldSelector):
-                for head, tail in other.heads.iteritems():
+                for head, tail in other.heads.items():
                     heads[head].add(tail)
             elif isinstance(other, FieldSelector):
                 selectors.append(other)
@@ -513,19 +595,20 @@ class MultiFieldSelector(object):
 
         self.heads = dict(
             (head, all if all in tail else MultiFieldSelector(*tail))
-            for head, tail in heads.iteritems()
+            for head, tail in heads.items()
         ) if None not in heads or heads[None] is not all else {None: all}
 
         # sanity assertions follow
         head_types = set(type(x) for x in self.heads)
-        self.has_int = int in head_types or long in head_types
-        self.has_string = any(issubclass(x, basestring) for x in head_types)
-        self.has_none = types.NoneType in head_types
+        self.has_int = int in head_types
+        self.has_string = str in head_types
+        self.has_none = type(None) in head_types
+        self.complete = self.has_none and self.heads[None] is all
         if self.has_none and (self.has_int or self.has_string):
             # this should be possible, but I'm punting on it for now
             raise ValueError(
                 "MultiFieldSelector cannot yet specify a list and a hash/"
-                "object at the same level: %r" % self.heads.keys()
+                "object at the same level: %r" % list(self.heads.keys())
             )
 
     def __str__(self):
@@ -543,14 +626,21 @@ class MultiFieldSelector(object):
 
     @property
     def path(self):
+        """The path attribute returns a stringified, concise representation of
+        the MultiFieldSelector.  It can be reversed by the ``from_path``
+        constructor.
+        """
         if len(self.heads) == 1:
-            return _fmt_mfs_path(self.heads.keys()[0], self.heads.values()[0])
+            return _fmt_mfs_path(
+                list(self.heads.keys())[0],
+                list(self.heads.values())[0]
+            )
         else:
             return "(" + "|".join(
-                _fmt_mfs_path(k, v) for (k, v) in self.heads.items()
+                _fmt_mfs_path(k, v) for (k, v) in list(self.heads.items())
             ) + ")"
 
-    def __nonzero__(self):
+    def __bool__(self):
         return bool(len(self.heads))
 
     def __iter__(self):
@@ -568,49 +658,88 @@ class MultiFieldSelector(object):
             <FieldSelector: .c>
             >>>
         """
-        for head, tail in self.heads.iteritems():
+        for head, tail in self.heads.items():
             head_selector = self.FieldSelector((head,))
             if tail is all:
                 if head is None:
-                    yield self.FieldSelector(())
-                yield head_selector
+                    yield self.FieldSelector((None,))
+                else:
+                    yield head_selector
             else:
                 for x in tail:
                     yield head_selector + x
 
     def __getitem__(self, index):
-        """Returns the MultiFieldSelector that applies to the specified
-        field/key/index.
+        """Returns the sub-MultiFieldSelector that applies to the specified
+        field/key/index.  If the MultiFieldSelector does not match any fields
+        at this path, it returns ``None``.  If it matches all fields at the
+        path, then it returns a 'complete' MultiFieldSelector (ie, a field
+        selector which matches everything).  If it matches a subset of fields
+        at that location, it will return a MultiFieldSelector which contains
+        those fields.
 
         ::
 
             >>> mfs = MultiFieldSelector(["a", "b"], ["a", "d"], ["c"])
             >>> mfs["a"]
             MultiFieldSelector(['b'], ['d'])
-            >>>
+            >>> mfs[("a", "c")]
+            None
         """
+        if self.complete:
+            return self
+
         if isinstance(index, (FieldSelector, tuple, list)):
+            # the semantics of this are a little different.  It should
+            # return a MultiFieldSelector, or None.
             if len(index) == 0:
                 return self
-            if self.has_none:
-                pass
-            elif len(index) == 1:
-                index = index[0]
+            head_index = None if self.has_none else index[0]
+            tail = index[1:]
+            if self.complete:
+                return self
+            elif head_index not in self.heads:
+                return None
+
+            head = self.heads[head_index]
+            if head is all:
+                return type(self).complete_mfs()
+            elif tail:
+                return head[tail]
             else:
-                return self.heads[index[0]][index[1:]]
+                return head
+
         if index is any:
+            # XXX useful?
             assert len(self.heads) <= 1, "ambigious fetch of 'any'"
             if len(self.heads) == 1:
-                index = self.heads.keys()[0]
+                index = list(self.heads.keys())[0]
             else:
-                return self
+                return self  # XXX wat
 
-        tail = self.heads[None] if self.has_none else self.heads[index]
-        return type(self)([None]) if tail == all else tail
+        tail = (
+            self.heads[None] if self.complete else
+            self.heads.get(index, None)
+        )
+        return type(self).complete_mfs() if tail == all else tail
 
     def __contains__(self, index):
         """Checks to see whether the given item matches the MultiFieldSelector.
 
+        For item index (string, number, ``None``) lookups, the result is True
+        if the result of applying the MultiFieldSelector (with ``.get``) would
+        return an item in that location (if it existed in the input).
+
+        For a FieldSelector (or sequence) lookups, the result is True if the
+        result of applying the MultiFieldSelector and then the FieldSelector on
+        an input is the same as applying the FieldSelector on the input.  ie,
+        that the MultiFieldSelector is a superset of the FieldSelector.
+
+        Unlike plain collections, use of ``in`` does not mean that subscripting
+        will return ``KeyError``.  When used with a FieldSelector, it returns
+        ``False`` when ``__getitem__`` with the index would not return a
+        complete FieldSelector, whereas the ``__getitem__`` call would return a
+        partial (non-complete) MultiFieldSelector.
         ::
 
             >>> mfs = MultiFieldSelector(["a", "b"], ["a", "d"], ["c"])
@@ -619,23 +748,19 @@ class MultiFieldSelector(object):
             >>> "b" in mfs
             False
             >>> FieldSelector(["a"]) in mfs
-            True
+            False
             >>> FieldSelector(["a", "d"]) in mfs
             True
             >>> FieldSelector(["a", "e"]) in mfs
             False
             >>>
         """
-        if isinstance(index, (FieldSelector, tuple, list)) and \
-                len(index) == 1:
-            index = index[0]
-
-        if isinstance(index, (basestring, types.IntType, types.NoneType)):
+        if isinstance(index, (str, int, type(None))):
             return self.has_none or index in self.heads
         elif index is any:
             return True if len(self.heads) else False
         elif len(index) == 0:
-            return True
+            return self.complete
         else:
             head_key = None if self.has_none else index[0]
             return (
@@ -680,22 +805,23 @@ class MultiFieldSelector(object):
             else:
                 vals = list(
                     self._get(obj[head], tail) for head, tail in
-                    self.heads.iteritems()
+                    sorted(self.heads.items())
                 )
+
             if isinstance(obj, ListCollection):
                 return ctor(values=vals)
             else:
                 return vals
-        elif isinstance(obj, dict):
+        elif isinstance(obj, (dict, DictCollection)):
             if self.has_none:
                 tail = self.heads[None]
                 return ctor(
-                    (k, self._get(v, tail)) for k, v in obj.iteritems()
+                    (k, self._get(v, tail)) for k, v in obj.items()
                 )
             else:
                 return ctor(
                     (head, self._get(obj[head], tail)) for head, tail in
-                    self.heads.iteritems() if head in obj
+                    self.heads.items() if head in obj
                 )
         else:
             if self.has_int or (self.has_none and self.heads[None] is not all):
@@ -708,7 +834,7 @@ class MultiFieldSelector(object):
                 return self._get(obj, all)
             else:
                 kwargs = dict()
-                for head, tail in self.heads.iteritems():
+                for head, tail in self.heads.items():
                     val = getattr(obj, head, None)
                     if val is not None:
                         kwargs[head] = self._get(val, tail)
@@ -754,7 +880,7 @@ class MultiFieldSelector(object):
             ``source=``\ *OBJECT*
                 the object to lift the fields from
 
-            ``copy=``\ *BOOL*\ |\ *FUNCTION*
+            ``copy=``\ *BOOL*\ \|\ *FUNCTION*
                 deep copy the values set, using copy.deepcopy (or the passed
                 function).  False by default.
         """
@@ -769,7 +895,7 @@ class MultiFieldSelector(object):
                 raise
 
         if copy and not callable(copy):
-            copy = lambda x: deepcopy(x)
+            copy = deepcopy
 
         for fs, val in fs_val:
             if val is _None:
@@ -784,6 +910,12 @@ class MultiFieldSelector(object):
         mfs = _scan_mfs_path(mfs_path)
         return cls(*mfs)
 
+    @classmethod
+    def complete_mfs(cls):
+        if not isinstance(cls._complete_mfs, cls):
+            cls._complete_mfs = cls([None])
+        return cls._complete_mfs
+
 
 _MFS_PATH_TOK = re.compile(
     r'''\.(?P<attr>\w+)|
@@ -799,9 +931,9 @@ _MFS_PATH_TOK = re.compile(
 
 
 def _scan_mfs_path(path):
-    mfs = list()
-    fs = list()
-    stack = list()
+    mfs = list()    # a list of parsed, complete FS's
+    fs = list()     # current cursor in parsing
+    stack = list()  # resume state when parens are closed
     popping = False
     for m in re.finditer(_MFS_PATH_TOK, path):
         if m.group("branch"):
@@ -835,4 +967,7 @@ def _scan_mfs_path(path):
             elif m.group('wild'):
                 fs.append(None)
 
+    if not popping and fs:
+        mfs.append(fs)
+
     return mfs
