@@ -16,8 +16,8 @@
 
 from __future__ import absolute_import
 
+from builtins import object
 import collections
-import types
 
 from normalize.coll import Collection
 import normalize.exc as exc
@@ -77,7 +77,7 @@ class Visitor(object):
 
         if visit_filter is None:
             visit_filter = filter
-        if isinstance(visit_filter, (MultiFieldSelector, types.NoneType)):
+        if isinstance(visit_filter, (MultiFieldSelector, type(None))):
             self.visit_filter = visit_filter
         else:
             self.visit_filter = MultiFieldSelector(*visit_filter)
@@ -87,8 +87,7 @@ class Visitor(object):
 
     def is_filtered(self, prop):
         return (not self.extraneous and prop.extraneous) or (
-            self.visit_filter and self.cue + [prop.name] not in
-            self.visit_filter
+            self.visit_filter and not self.visit_filter[self.cue + [prop.name]]
         )
 
     @property
@@ -214,7 +213,9 @@ class VisitorPattern(object):
                 This function should take a
                 :py:class:`normalize.property.Property` instance, and return
                 the slot from the value, or raise ``AttributeError`` or
-                ``KeyError`` if the slot is empty.
+                ``KeyError`` if the slot is empty.  Returning nothing means
+                that the item has no properties to unpack; ie, it's an opaque
+                type.
 
             ``get_item=``\ *generator*
                 This generator should return the tuple protocol used by
@@ -230,14 +231,20 @@ class VisitorPattern(object):
                     generator = value_type.coll_to_tuples(value)
                 else:
                     raise exc.VisitorUnpackError(
-                        value=repr(value),
+                        passed=value,
                         colltype=value_type.colltype.__name__,
-                        fs=visitor.field_selector,
+                        context=visitor,
                     )
         else:
             generator = None
 
-        return (lambda prop: prop.__get__(value)), generator
+        if issubclass(value_type, Record):
+            def propget(prop):
+                return prop.__get__(value)
+        else:
+            propget = None
+
+        return propget, generator
 
     @classmethod
     def apply(cls, value, prop, visitor):
@@ -262,6 +269,9 @@ class VisitorPattern(object):
                 This is the :py:class:`normalize.Property` instance which
                 represents the field being traversed.
 
+                This can be ``None`` when being applied over Collection
+                instances, where the type of the contents is not a Record.
+
             ``visitor=``\ *Visitor*
                 This object can be used to inspect parameters of the current
                 run, such as options which control which kinds of values are
@@ -334,7 +344,7 @@ class VisitorPattern(object):
 
         if issubclass(value_type, Collection) and aggregated is not None:
             if all(visitor.is_filtered(prop) for prop in
-                   value_type.properties.values()):
+                   list(value_type.properties.values())):
                 reduced = aggregated
             else:
                 if reduced.get("values", False):
@@ -379,8 +389,8 @@ class VisitorPattern(object):
     # note: Collection.coll_to_tuples will generally allow you to pass
     # collections as a list or a dict with the *values* being the members of
     # the set, so this code allows this.
-    grok_mapping_types = collections.Mapping
-    grok_coll_types = (collections.Sequence, collections.Mapping)
+    grok_mapping_types = collections.abc.Mapping
+    grok_coll_types = (collections.abc.Sequence, collections.abc.Mapping)
 
     @classmethod
     def grok(cls, value, value_type, visitor):
@@ -397,8 +407,10 @@ class VisitorPattern(object):
         ``value_type`` is not appropriate for ``value``.
         """
         is_coll = issubclass(value_type, Collection)
-        is_record = any(not visitor.is_filtered(prop) for prop in
-                        value_type.properties.values())
+        is_record = issubclass(value_type, Record) and any(
+            not visitor.is_filtered(prop) for prop in
+            list(value_type.properties.values())
+        )
 
         if is_record and not isinstance(value, cls.grok_mapping_types):
             raise exc.VisitorGrokRecordError(
@@ -427,7 +439,12 @@ class VisitorPattern(object):
                 )
             generator = value_type.coll_to_tuples(values)
 
-        return (lambda prop: value[prop.name]), generator
+        propget = None
+        if is_record:
+            def propget(prop):
+                return value[prop.name]
+
+        return propget, generator
 
     @classmethod
     def reverse(cls, value, prop, visitor):
@@ -508,7 +525,12 @@ class VisitorPattern(object):
 
             item_type_generator = get_item_types()
 
-        return (lambda prop: prop), item_type_generator
+        propget = None
+        if issubclass(value_type, Record):
+            def propget(prop):
+                return prop
+
+        return propget, item_type_generator
 
     @classmethod
     def propinfo(cls, value, prop, visitor):
@@ -517,6 +539,9 @@ class VisitorPattern(object):
         implementation returns just the name of the property and the type in
         here.
         """
+        if not prop:
+            return {"name": value.__name__}
+
         rv = {"name": prop.name}
         if prop.valuetype:
             if isinstance(prop.valuetype, tuple):
@@ -606,6 +631,8 @@ class VisitorPattern(object):
         mapped_props = None
         if props:
             mapped_props = cls.map_record(visitor, props, value_type)
+        elif mapped_coll is None:
+            return visitor.apply(value, None, visitor)
 
         return visitor.reduce(
             mapped_props, mapped_coll, value_type, visitor,
@@ -614,7 +641,7 @@ class VisitorPattern(object):
     @classmethod
     def map_record(cls, visitor, get_value, record_type):
         rv = visitor.copy()  # expensive?
-        for name, prop in record_type.properties.iteritems():
+        for name, prop in record_type.properties.items():
             if rv.is_filtered(prop):
                 continue
 
